#include "utils.h"
#include "layout.h"
#include "wingedmonkey.h"
#include "cmdeng.h"
#include "sock.h"

typedef struct {
  const char* string;
  void (*fun)(int, char**);
} funasoc;

static const funasoc translation[]={
  {"client-get",cmd_client_get},
  {"client-set",cmd_client_set},
  {"client-count",cmd_client_count},
  {"client-focus",cmd_client_focus},
  {"client-kill",cmd_client_kill},
  {"client-list",cmd_client_list},
  {"client-list-stack",cmd_client_list_stack},
  {"client-move",cmd_client_move},
  {"client-find",cmd_client_find},
  {"desktop-get",cmd_desktop_get},
  {"desktop-set",cmd_desktop_set},
  {"desktop-add",cmd_desktop_add},
  {"desktop-list",cmd_desktop_list},
  {"desktop-count",cmd_desktop_count},
  {"desktop-next",cmd_desktop_next},
  {"desktop-prev",cmd_desktop_prev},
  {"desktop-remove",cmd_desktop_remove},
  {"desktop-state",cmd_desktop_state},
  {"desktop-change",cmd_desktop_change},
  {"desktop-rename",cmd_desktop_rename},
  {"layer-get",cmd_layer_get},
  {"layer-set",cmd_layer_set},
  {"layer-add",cmd_layer_add},
  {"layer-list",cmd_layer_list},
  {"layer-list-stack",cmd_layer_list_stack},
  {"layer-count",cmd_layer_count},
  {"layer-remove",cmd_layer_remove},
  {"tile-get",cmd_tile_get},
  {"tile-set",cmd_tile_set},
  {"tile-add",cmd_tile_add},
  {"tile-list",cmd_tile_list},
  {"tile-remove",cmd_tile_remove},
  {"rule-add",cmd_rule_add},
  {"rule-remove",cmd_rule_remove},
  {"rule-list",cmd_rule_list},
  {"display-get",cmd_display_get},
  {"hook-set",cmd_hook_set},
  {"hook-list",cmd_hook_list},
  {"cycle",cmd_cycle},
  {"quit",cmd_quit},
  {"exec",cmd_exec},
  {"include",cmd_include},
  {"help",cmd_help}
};

static void decode_xy(char* str, int* va, float* vr){
  char* c=strchr(str,'%');
  if(!c){
    *va=strtol(str,0,10);
    *vr=0;
  }else{
    *va=strtol(c+1,0,10);
    *c=0;
    *vr=strtod(str,0)/100;
    *c='%';
  }
}

static int get_xy(char* str, int full){
  char* c=strchr(str,'%');
  if(!c){
    return strtol(str,0,10);
  }else{
    *c=0;
    int result=strtol(c+1,0,10)+(int)(strtod(str,0)/100*full);
    *c='%';
    return result;
  }
}


Layer* decode_layer(char* name){
  if(!name[0]){
    if(MONITOR->focus){
      return MONITOR->focus->layer;
    }else{
      return 0;
    }
  }
  if(isalpha(name[0])){
    return str_to_layer(MONITOR,name);
  }else{
    int id=strtol(name,0,10);
    if(id>=0){
      return id_to_layer(MONITOR,id);
    }
  }
  return 0;
}

Tile* decode_tile(char* name){
  if(!name[0]){
    if(MONITOR->focus){
      return MONITOR->focus->tile;
    }else{
      return 0;
    }
  }

  Tile* t=0;

  char* dot=strchr(name,'.');
  if(dot){
    dot[0]=0;
    Layer* l=decode_layer(name);
    if(!l)return 0;
    if(!dot[1]){
      if(MONITOR->focus){
	t=MONITOR->focus->tile;
      }
    }else if(isalpha(dot[1])){
      t=str_to_ltile(l,dot+1);
    }else{
      int id=strtol(dot+1,0,10);
      if(id>=0){
	t=id_to_ltile(l,id);
      }
    }
    dot[0]='.';
  }else{
    t=str_to_tile(MONITOR,name);
  }

  return t;
}

Client* decode_client(char* name){
  if(!strcmp(name,"FOCUS")){
    return MONITOR->focus;
  }
  if(!strcmp(name,"focus")){
    return MONITOR->focus;
  }
  int win=strtol(name,0,0);
  return win_to_client(MONITOR,(Window)win);
}

Dsktp* decode_desktop(char* name){
  if(isalpha(name[0])){
    return str_to_desktop(MONITOR, name);
  }else{
    int id=strtol(name,0,10);
    return id_to_desktop(MONITOR,id+1);
  }
  return 0;
}

const char* stacklvl_strings[]={
  "desktop",
  "bottom",
  "normal",
  "top",
  "dock"
};

const char* floattype_strings[]={
  "floating",
  "maximized",
  "fullscreen"
};

const char* sticky_strings[]={
  "normal",
  "sticky",
  "slippery",
  "avoid"
};

const char* boolean_strings[]={
  "no","yes"
};



void cmd_client_get(int argc, char** argz){
  if(argc==0){
    sockreport("0x%x",MONITOR->focus->win);
    return;
  }
  Client* c=decode_client(argz[0]);
  if(!c){
    sockreport("invalid client");
    return;
  }
  if(argc==1 || !strcmp(argz[1],"name"))
    sockreport("name\t%s",c->name);
  if(argc==1 || !strcmp(argz[1],"role"))
    sockreport("role\t%s",c->role);
  if(argc==1 || !strcmp(argz[1],"class")){
    sockreport("classname\t%s",c->classN);
    sockreport("class\t%s",c->classC);
  }
  if(argc==1 || !strcmp(argz[1],"geometry"))
    sockreport("geometry\t%d %d %d %d",c->x,c->y,c->w,c->h);
  if(argc==1 || !strcmp(argz[1],"stack-level"))
    sockreport("stack-level\t%s",stacklvl_strings[c->stacktype]);
  if(argc==1 || !strcmp(argz[1],"float-type"))
    sockreport("float-type\t%s",floattype_strings[c->float_state]);
  if(argc==1 || !strcmp(argz[1],"size-hints"))
    sockreport("size-hints\t%d %d %d %d %d %d %f %f",
	c->basew,c->baseh,c->incw,c->inch,
	c->maxw,c->maxh,c->mina,c->maxa
	);
  if(argc==1 || !strcmp(argz[1],"visible"))
    sockreport("visible\t%s",boolean_strings[c->visible]);
  if(argc==1 || !strcmp(argz[1],"layer"))
    sockreport("layer\t%s",c->layer->name);
  if(argc==1 || !strcmp(argz[1],"tile"))
    sockreport("tile\t%s.%s",c->layer->name,c->tile->name);
  if(argc==1 || !strcmp(argz[1],"transients")){
    for(Client* chld=c->children;chld;chld=chld->chld_next){
      sockreport("0x%x",chld->win);
    }
  }
  return;
}

void cmd_client_set(int argc, char** argz){
  if(argc<2){
    return;
  }
  Client* c=decode_client(argz[0]);
  if(!c){
    sockreport("invalid client");
    return;
  }
  if(!strcmp(argz[1],"geometry")){
    if(argc!=4 && argc!=6)return;
    if(strcmp("-",argz[2])){
      if(argz[2][0]=='t'){
	c->x=c->tile->x+get_xy(argz[2]+1,c->tile->w);
      }else {
	c->x=get_xy(argz[2],MONITOR->mw);
      }
    }
    if(strcmp("-",argz[3])){
      if(argz[3][0]=='t'){
	c->y=c->tile->y+get_xy(argz[3]+1,c->tile->h);
      }else {
	c->y=get_xy(argz[3],MONITOR->mh);
      }
    }
    if(argc==6){
      if(strcmp("-",argz[4])){
	if(argz[4][0]=='t'){
	  c->w=get_xy(argz[4]+1,c->tile->w);
	}else{
	  c->w=get_xy(argz[4],MONITOR->mw);
	}
      }
      if(strcmp("-",argz[5])){
	if(argz[5][0]=='t'){
	  c->h=get_xy(argz[5]+1,c->tile->h);
	}else{
	  c->h=get_xy(argz[5],MONITOR->mh);
	}
      }
    }
    client_reshape(c,0);
    xclient_moveresize(c);
    return;
  }
  if(!strcmp(argz[1],"stack-level")){
    if(argc!=3)return;
    for(int i=StckDesktop;i<=StckDock;i++){
      if(!strcmp(argz[2],stacklvl_strings[i])){
	xclient_stackrel(c,i);
	return;
      }
    }
    sockreport("unknown stack level");
    return;
  }
  if(!strcmp(argz[1],"float-type")){
    if(argc!=3)return;
    for(int i=FloatFloat;i<=FloatFull;i++){
      if(!strcmp(argz[2],floattype_strings[i])){
	xclient_floatrel(c,i);
	return;
      }
    }
    sockreport("unknown float type");
    return;
  }

}

void cmd_client_count(int argc, char** argz){
  if(argc==0){
    sockreport("%d",MONITOR->clientcount);
    return;
  }
  if(strchr(argz[0],'.')){
    Tile* t=decode_tile(argz[0]);
    if(!t){
      sockreport("invalid tile");
      return;
    }
    sockreport("%d",t->clientcount);
  }else{
    Layer* l=decode_layer(argz[0]);
    if(!l){
      sockreport("invalid layer");
      return;
    }
    sockreport("%d",l->clientcount);
  }
  return;
}

void cmd_client_focus(int argc, char** argz){
  if(argc==0){
    return;
  }
  Client* c=decode_client(argz[0]);
  if(!c){
    sockreport("invalid client");
    return;
  }
  if(!c->visible){
    sockreport("cannot focus invisible client");
    return;
  }
  xclient_activate(c);
  return;
}

void cmd_client_kill(int argc, char** argz){
  if(argc==0){
    return;
  }
  Client* c=decode_client(argz[0]);
  if(!c){
    sockreport("invalid client");
    return;
  }
  int arg=0;
  if(argc>1){
    arg=strtol(argz[1],0,10);
    if(arg<0 || arg>64)arg=0;
  }
  if(arg==0){
    xclient_kill(c);
  }else{
    xclient_murder(c,arg);
  }
  return;
}

void cmd_client_list(int argc, char** argz){
  if(argc==0){
    for(Client* c=MONITOR->firstclient;c;c=c->next){
      sockreport("0x%x",c->win);
    }
    return;
  }
  if(strchr(argz[0],'.')){
    Tile* t=decode_tile(argz[0]);
    if(!t){
      sockreport("invalid tile");
      return;
    }
    for(Client* c=t->firstclient;c;c=c->snext){
      sockreport("0x%x",c->win);
    }
  }else{
    Layer* l=decode_layer(argz[0]);
    if(!l){
      sockreport("invalid layer");
      return;
    }
    for(Tile* t=l->firsttile;t;t=t->next){
      for(Client* c=t->firstclient;c;c=c->snext){
	sockreport("0x%x",c->win);
      }
    }
  }
  return;
}

void cmd_client_list_stack(int argc, char** argz){
  for(Client* c=MONITOR->firstfocus;c;c=c->fnext){
    if(c->visible){
      sockreport("0x%x\t%s\tvisible",c->win,c->name);
    }else{
      sockreport("0x%x\t%s\thidden",c->win,c->name);
    }
  }
  return;
}

void cmd_client_move(int argc, char** argz){
  if(argc<2){
    return;
  }
  Client* c=decode_client(argz[0]);
  if(!c){
    sockreport("invalid client");
    return;
  }
  if(!strcmp(argz[1],"+")){//to next tile
    Tile* t=c->tile;
    do { t=tile_cycle(t); } while(t!=c->tile && !tile_is_free(t));
    if(t==c->tile)return;
    client_move_to_tile(t,MONITOR->focus);
    rect_findbest(c,c->tile);
    client_reshape(c,0);
    xclient_moveresize(c);
    xclient_configure(c,CWBorderWidth);
    xorg_restack();
    return;
  }
  if(strchr(argz[1],'.')){
    Tile* t=decode_tile(argz[1]);
    if(!t){
      sockreport("invalid tile");
      return;
    }
    client_move_to_tile(t,c);
    rect_findbest(c,t);
    client_reshape(c,0);
    xclient_moveresize(c);
    xclient_configure(c,CWBorderWidth);
    xorg_restack();
  }else{
    Layer* l=decode_layer(argz[1]);
    if(!l){
      sockreport("invalid layer");
      return;
    }
    xclient_changelayer(c,l);
    xorg_refreshfocus();
  }
  return;
}

void cmd_client_find(int argc, char** argz){
  if(argc<1)return;
  char* str1=0;
  char* str2=0;
  char* str3=0;
  char* str4=0;
  for(int i=0;i<argc;i++){
    char* eq=strchr(argz[i],'=');
    if(!eq)continue;
    eq[0]=0;
    if(!strcmp(argz[i],"name")){str1=eq+1;}else
    if(!strcmp(argz[i],"role")){str2=eq+1;}else
    if(!strcmp(argz[i],"classname")){str3=eq+1;}else
    if(!strcmp(argz[i],"class")){str4=eq+1;}else
      sockreport("unknown property to match: %s",argz[i]);

    eq[0]='=';
  }
  if((str1) && (*str1==0))str1=0;
  if((str2) && (*str2==0))str2=0;
  if((str3) && (*str3==0))str3=0;
  if((str4) && (*str4==0))str4=0;
  for(Client* c=MONITOR->firstclient;c;c=c->next){
    int passed=1;
    if(str1){
      passed=passed&&(!strcmp(str1,c->name));
    }
    if(str2){
      passed=passed&&(!strcmp(str2,c->role));
    }
    if(str3){
      passed=passed&&(!strcmp(str3,c->classN));
    }
    if(str4){
      passed=passed&&(!strcmp(str4,c->classC));
    }
    if(passed){
      sockreport("0x%x\t%s",c->win,c->name);
    }
  }
  return;
}


void cmd_desktop_set(int argc, char** argz){
  if(argc<1)return;
  if(argz[0][0]=='!'){
    for(char* c=argz[0]+1;*c;c++){
      if(*c!='*' && *c!='+' && *c!='0' && *c!='X'){
	sockreport("invalid desktop layout");
	return;
      }
    }
    xlayer_applylayout(argz[0]+1,0);
  }else{
    Dsktp* d=decode_desktop(argz[0]);
    if(!d){
      sockreport("invalid desktop");
      return;
    }
    xdesktop_setdesktop(d);
  }
  return;
}

void cmd_desktop_get(int argc, char** argz){
  Dsktp* d=0;
  if(argc==0){
    d=MONITOR->desktop;
  }else{
    d=decode_desktop(argz[0]);
  }
  if(!d){
    sockreport("invalid desktop");
    return;
  }
  sockreport("%s\t%.*s",d->name,MONITOR->layercount,d->layout);
  return;
}

void cmd_desktop_add(int argc, char** argz){
  if(argc<2)return;
  for(char* c=argz[1];*c;c++){
    if(*c!='*' && *c!='+' && *c!='0' && *c!='X'){
      sockreport("invalid desktop layout");
      return;
    }
  }
  Dsktp* d=new_desktop(argz[0],argz[1]);
  desktop_add(MONITOR,d);
  if(!MONITOR->desktop){
    xdesktop_setdesktop(d);
  }
  xorg_refreshalldesktop();
  xorg_desktopsupdate();
  return;
}

void cmd_desktop_list(int argc, char** argz){
  for(Dsktp* d=MONITOR->firstdesktop;d;d=d->next){
    sockreport("%s\t%.*s",d->name,MONITOR->layercount,d->layout);
  }
  return;
}

void cmd_desktop_count(int argc, char** argz){
  sockreport("%d",MONITOR->desktopcount);
  return;
}

void cmd_desktop_next(int argc, char** argz){
  user_desktopnext(1);
  return;
}

void cmd_desktop_prev(int argc, char** argz){
  user_desktopnext(-1);
  return;
}

void cmd_desktop_remove(int argc, char** argz){
  if(argc<1)return;
  Dsktp* d=decode_desktop(argz[0]);
  if(!d){
    sockreport("invalid desktop");
    return;
  }
  if(d==MONITOR->desktop){
    sockreport("warning: removing active desktop");
    MONITOR->desktop=0;
    desktop_remove(d);
    xdesktop_setdesktop(MONITOR->firstdesktop);
  }else{
    desktop_remove(d);
  }
  xorg_refreshalldesktop();
  xorg_desktopsupdate();
  free(d);
}

void cmd_desktop_state(int argc, char** argz){
  char* dmask=malloc(MONITOR->layercount+1);
  int i=0;
  for(Layer* l=MONITOR->firstlayer;l;l=l->next){
    dmask[i]=l->visible?'X':'0';
    i++;
  }
  dmask[i]=0;
  sockreport("%s",dmask);
  free(dmask);
}


void cmd_desktop_change(int argc, char** argz){
  if(argc<2)return;
  for(char* c=argz[1];*c;c++){
    if(*c!='*' && *c!='+' && *c!='0' && *c!='X'){
      sockreport("invalid desktop layout");
      return;
    }
  }
  Dsktp* d=decode_desktop(argz[0]);
  Dsktp* d2=new_desktop(d->name,argz[1]);
  d2->prev=d->prev;
  d2->next=d->next;
  d2->mon=d->mon;
  d2->id=d->id;
  if(d2->prev)d2->prev->next=d2;
  if(d2->next)d2->next->prev=d2;
  if(d==d->mon->lastdesktop)d->mon->lastdesktop=d2;
  if(d==d->mon->firstdesktop)d->mon->firstdesktop=d2;
  if(d==d->mon->desktop){
    d->mon->desktop=d2;
    xdesktop_setdesktop(d2);
  }
  free(d);
  xorg_refreshalldesktop();
  xorg_desktopsupdate();
  return;
}


void cmd_desktop_rename(int argc, char** argz){
  if(argc!=2)return;
  Dsktp* d=decode_desktop(argz[0]);
  if(!d){
    sockreport("invalid desktop");
    return;
  }
  strcpy(d->name,argz[1]);
  xorg_desktopsupdate();
  return;
}


void cmd_layer_get(int argc, char** argz){
  if(argc==0){
    sockreport("%s",MONITOR->sfirstlayer->name);
    return;
  }
  Layer* l=decode_layer(argz[0]);
  if(!l){
    sockreport("invalid layer");
    return;
  }
  if(argc==1 || !strcmp(argz[1],"name"))
    sockreport("name\t%s",l->name);
  if(argc==1 || !strcmp(argz[1],"index"))
    sockreport("index\t%d",layer_to_id(MONITOR,l));
  if(argc==1 || !strcmp(argz[1],"focus-or-hide"))
    sockreport("focus-or-hide\t%s",boolean_strings[l->focusorhide]);
  if(argc==1 || !strcmp(argz[1],"stack-level"))
    sockreport("stack-level\t%s",stacklvl_strings[l->stacktype]);
  if(argc==1 || !strcmp(argz[1],"visible"))
    sockreport("visible\t%s",boolean_strings[l->visible]);
  return;

}

void cmd_layer_set(int argc, char** argz){
  if(argc<2){
    return;
  }
  Layer* l=decode_layer(argz[0]);
  if(!l){
    sockreport("invalid layer");
    return;
  }
  if(!strcmp(argz[1],"focus-or-hide")){
    if(argc!=3)return;
    int oldstate=l->focusorhide;
    int newstate=!oldstate;
    if(!strcmp(argz[2],"yes")){
      newstate=1;
    }else if(!strcmp(argz[2],"no")){
      newstate=0;
    }
    if(oldstate!=newstate){
      l->focusorhide=newstate;
      if(newstate){
	if(!l->sprev){
	  xlayer_visible(l,0);
	  xorg_refreshfocus();
	}
      }
    }
    return;
  }
  if(!strcmp(argz[1],"stack-level")){
    if(argc!=3)return;
    for(int i=StckBottom;i<=StckTop;i++){
      if(!strcmp(argz[2],stacklvl_strings[i])){
	l->stacktype=i;
	xorg_restack();
	return;
      }
    }
    sockreport("unknown stack level");
    return;
  }
  if(!strcmp(argz[1],"visible")){
    if(argc!=3)return;
    int oldstate=l->visible;
    int newstate=!oldstate;
    if(!strcmp(argz[2],"yes")){
      newstate=1;
    }else if(!strcmp(argz[2],"no")){
      newstate=0;
    }
    if(oldstate!=newstate){
      user_togglelayer(layer_to_id(MONITOR,l));
    }
    return;
  }
  if(!strcmp(argz[1],"name")){
    if(argc!=3)return;
    strcpy(l->name,argz[2]);
    return;
  }
}

void cmd_layer_list(int argc, char** argz){
  for(Layer* l=MONITOR->firstlayer;l;l=l->next){
    sockreport("%s",l->name);
  }
  return;
}

void cmd_layer_list_stack(int argc, char** argz){
  for(Layer* l=MONITOR->sfirstlayer;l;l=l->snext){
    if(l->visible){
      sockreport("%s\tvisible",l->name);
    }else{
      sockreport("%s\thidden",l->name);
    }
  }
  return;
}

void cmd_layer_count(int argc, char** argz){
  sockreport("%d",MONITOR->layercount);
  return;
}

void cmd_layer_add(int argc, char** argz){
  if(argc!=1)return;
  Layer* l=new_layer(argz[0]);
  layer_add(MONITOR,l);
  layer_stack_add_back(MONITOR,l);
  xdesktop_setdesktop(MONITOR->desktop);
  return;
}

void cmd_layer_remove(int argc, char** argz){
  if(argc!=1)return;
  Layer* l=decode_layer(argz[0]);
  if(!l){
    sockreport("invalid layer");
    return;
  }
  if(!layer_to_id(MONITOR,l)){
    sockreport("cannot remove default (zeroth) layer");
    return;
  }
  layer_stack_remove(l);
  layer_remove(l);
  Tile* t=0;
  while((t=l->firsttile)){
    Client* c=0;
    while((c=t->firstclient)){
      client_stack_remove(c);
      client_stack_add_to_monitor(MONITOR,c);
      client_reshape(c,0);
      xclient_moveresize(c);
    }
    tile_remove(t);
    free(t);
  }
  free(l);
  xorg_refreshfocus();
  xorg_restack();
  xdesktop_setdesktop(MONITOR->desktop);
  return;
}

void cmd_tile_get(int argc, char** argz){
  if(argc<1)return;
  Tile* t=decode_tile(argz[0]);
  if(!t){
    sockreport("invalid tile");
    return;
  }
  if(argc==1 || !strcmp(argz[1],"name"))
    sockreport("name\t%s",t->name);
  if(argc==1 || !strcmp(argz[1],"layer"))
    sockreport("layer\t%s",t->layer->name);
  if(argc==1 || !strcmp(argz[1],"index")){
    int index=0;
    for(Tile* tx=t->layer->firsttile;tx;tx=tx->next){
      if(tx==t)break;
      index++;
    }
    sockreport("index\t%d",index);
  }
  if(argc==1 || !strcmp(argz[1],"geometry"))
    sockreport("geometry\t%d %d %d %d",t->x,t->y,t->w,t->h);
  if(argc==1 || !strcmp(argz[1],"force-maximize"))
    sockreport("force-maximize\t%s",boolean_strings[t->force_maximize]);
  if(argc==1 || !strcmp(argz[1],"sticky"))
    sockreport("sticky\t%s",sticky_strings[t->sticky]);
  if(argc==1 || !strcmp(argz[1],"max-clients"))
    sockreport("max-clients\t%d",t->max_clients);
  if(argc==1 || !strcmp(argz[1],"skip-focus"))
    sockreport("skip-focus\t%s",boolean_strings[t->skip_focus]);
  return;
}

void cmd_tile_set(int argc, char** argz){
  if(argc<2)return;
  Tile* t=decode_tile(argz[0]);
  if(!t){
    sockreport("invalid tile");
    return;
  }
  if(!strcmp(argz[1],"geometry")){
    if(argc!=4 && argc!=6)return;
    if(strcmp("-",argz[2]))decode_xy(argz[2],&(t->xa),&(t->xr));
    if(strcmp("-",argz[3]))decode_xy(argz[3],&(t->ya),&(t->yr));
    if(argc==6){
      if(strcmp("-",argz[4]))decode_xy(argz[4],&(t->wa),&(t->wr));
      if(strcmp("-",argz[5]))decode_xy(argz[5],&(t->ha),&(t->hr));
    }
    tile_recalc(t);
    for(Client* c=t->firstclient;c;c=c->snext){
      client_reshape(c,0);
      xclient_moveresize(c);
    }
    return;
  }
  if(!strcmp(argz[1],"force-maximize")){
    if(argc!=3)return;
    int oldstate=t->force_maximize;
    int newstate=!oldstate;
    if(!strcmp(argz[2],"yes")){
      newstate=1;
    }else if(!strcmp(argz[2],"no")){
      newstate=0;
    }
    if(oldstate!=newstate){
      t->force_maximize=newstate;
      for(Client* c=t->firstclient;c;c=c->snext){
	client_reshape(c,0);
	xclient_moveresize(c);
      }
    }
    return;
  }
  if(!strcmp(argz[1],"skip-focus")){
    if(argc!=3)return;
    int oldstate=t->skip_focus;
    int newstate=!oldstate;
    if(!strcmp(argz[2],"yes")){
      newstate=1;
    }else if(!strcmp(argz[2],"no")){
      newstate=0;
    }
    if(oldstate!=newstate){
      t->skip_focus=newstate;
    }
    return;
  }
  if(!strcmp(argz[1],"sticky")){
    if(argc!=3)return;
    for(int i=StickNormal;i<=StickSlippery;i++){
      if(!strcmp(argz[2],sticky_strings[i])){
	t->sticky=i;
	return;
      }
    }
    sockreport("unknown sticky specification");
    return;
  }
  if(!strcmp(argz[1],"max-clients")){
    if(argc!=3)return;
    int newmax=strtol(argz[2],0,10);
    if(newmax<0){
      sockreport("negative max-clients forbidden");
      return;
    }
    t->max_clients=newmax;
    return;
  }
  if(!strcmp(argz[1],"name")){
    if(argc!=3)return;
    strcpy(t->name,argz[2]);
    return;
  }
  return;
}

void cmd_tile_add(int argc, char** argz){
  if(argc!=5)return;
  char* dot=strchr(argz[0],'.');
  if(dot){
    dot[0]=0;
    Layer* l=decode_layer(argz[0]);
    if(!l){
      sockreport("invalid layer");
    }else if(!l->prev){
      sockreport("zeroth layer is read-only");
    }else{
      sockreport("tile-adding %s",dot+1);
      int xa,ya,wa,ha;
      float xr,yr,wr,hr;
      decode_xy(argz[1],&xa,&xr);
      decode_xy(argz[2],&ya,&yr);
      decode_xy(argz[3],&wa,&wr);
      decode_xy(argz[4],&ha,&hr);
      Tile* t=new_tile_full(dot+1,xa,ya,wa,ha,xr,yr,wr,hr);

      tile_add(l,t);
    }
    dot[0]='.';
  }else{
    sockreport("added tile must have specified layer: \"layer.tile\"");
  }
  return;
}


void cmd_tile_list(int argc, char** argz){
  if(argc==0){
    for(Layer* l=MONITOR->firstlayer;l;l=l->next){
      for(Tile* t=l->firsttile;t;t=t->next){
	sockreport("%s.%s",l->name,t->name);
      }
    }
  }else{
    Layer* l=decode_layer(argz[0]);
    if(!l){
      sockreport("invalid layer");
    return;
    }
    for(Tile* t=l->firsttile;t;t=t->next){
      sockreport("%s.%s",l->name,t->name);
    }
  }
  return;
}


void cmd_tile_remove(int argc, char** argz){
  if(argc!=1)return;
  Tile* t=decode_tile(argz[0]);
  if(!t){
    sockreport("invalid tile");
    return;
  }
  if(!layer_to_id(MONITOR,t->layer)){
    sockreport("cannot remove tile from default (zeroth) layer");
    return;
  }

  t->max_clients=1;
  Client* cnext=0;
  for(Client* c=t->firstclient;c;c=cnext){
    cnext=c->snext;
    client_stack_remove(c);
    //it's empty, so we can remove the tile
    if(!cnext)tile_remove(t);
    client_stack_add_to_monitor(MONITOR,c);
    client_reshape(c,0);
    xclient_moveresize(c);
  }
  tile_remove(t);
  free(t);
  xorg_refreshfocus();
  xorg_restack();
  return;
}

void cmd_rule_add(int argc, char** argz){
  if(argc<1)return;
  char* str1=0;
  char* str2=0;
  char* str3=0;
  char* str4=0;
  char* strx=0;
  char* stry=0;
  char* strw=0;
  char* strh=0;
  char* strv=0;
  char* strfl=0;
  char* strsl=0;
  char* strst=0;

  int i;
  for(i=0;i<argc;i++){
    if(!strcmp(argz[i],"--"))break;
    char* eq=strchr(argz[i],'=');
    if(!eq)continue;
    eq[0]=0;
    if(!strcmp(argz[i],"name")){str1=eq+1;}else
    if(!strcmp(argz[i],"role")){str2=eq+1;}else
    if(!strcmp(argz[i],"classname")){str3=eq+1;}else
    if(!strcmp(argz[i],"class")){str4=eq+1;}else
      sockreport("unknown property to match: %s",argz[i]);

    eq[0]='=';
  }
  char* tilename=0;
  for(i++;i<argc;i++){
    char* eq=strchr(argz[i],'=');
    if(!eq)continue;
    eq[0]=0;
    if(!strcmp(argz[i],"tile")){tilename=eq+1;}else
    if(!strcmp(argz[i],"x")){strx=eq+1;}else
    if(!strcmp(argz[i],"y")){stry=eq+1;}else
    if(!strcmp(argz[i],"w")){strw=eq+1;}else
    if(!strcmp(argz[i],"h")){strh=eq+1;}else
    if(!strcmp(argz[i],"vo")){strv=eq+1;}else
    if(!strcmp(argz[i],"float-type")){strfl=eq+1;}else
    if(!strcmp(argz[i],"stack-level")){strsl=eq+1;}else
    if(!strcmp(argz[i],"no-steal")){strst=eq+1;}else
      sockreport("unknown property to set: %s",argz[i]);

    eq[0]='=';
  }
  if(str1 && *str1==0)str1=0;
  if(str2 && *str2==0)str2=0;
  if(str3 && *str3==0)str3=0;
  if(str4 && *str4==0)str4=0;
  Tile* t=0;
  if(tilename)t=decode_tile(tilename);
  /*
  if(!t){
    sockreport("invalid tile");
    return;
  }
  */
  Rule* r=add_rule(MONITOR,str1,str2,str3,str4,t);
  if(strx && *strx!=0)decode_xy(strx,&r->x,&r->xr);
  if(stry && *stry!=0)decode_xy(stry,&r->y,&r->yr);
  if(strw && *strw!=0)decode_xy(strw,&r->w,&r->wr);
  if(strh && *strh!=0)decode_xy(strh,&r->h,&r->hr);
  if(strv && *strv!=0)if(!strcmp(strv,"yes"))r->visible_only=1;
  if(strst && *strst!=0)if(!strcmp(strst,"yes"))r->no_steal=1;
  if(strfl && *strfl!=0){
    for(int i=FloatFloat;i<=FloatFull;i++){
      if(!strcmp(strfl,floattype_strings[i])){
	r->float_state=i;
	break;
      }
    }
  }
  if(strsl && *strsl!=0){
    for(int i=StckDesktop;i<=StckDock;i++){
      if(!strcmp(strsl,stacklvl_strings[i])){
	r->stack_level=i;
	break;
      }
    }
  }
  return;
}

void cmd_rule_remove(int argc, char** argz){
  if(argc<1)return;
  int todelete=strtol(argz[0],0,10);
  int index=0;
  for(Rule* r=MONITOR->rulz;r;r=r->next){
    if(index==todelete){
      remove_rule(MONITOR,r);
      return;
    }
    ++index;
  }
  sockreport("rule index out of bounds: %d",todelete);
  return;
}

void cmd_rule_list(int argc, char** argz){
  char buffer[256];
  int i=0;
  for(Rule* r=MONITOR->rulz;r;r=r->next){//what about the geometry?
    char* pos=buffer;
    pos+=sprintf(pos,"[%d]",i);
    if(r->name)pos+=sprintf(pos,"\tname=%s",r->name);
    if(r->role)pos+=sprintf(pos,"\trole=%s",r->role);
    if(r->classN)pos+=sprintf(pos,"\tclassname=%s",r->classN);
    if(r->classC)pos+=sprintf(pos,"\tclass=%s",r->classC);
    pos+=sprintf(pos,"\t--");
    if(r->target)pos+=sprintf(pos,"\ttile=%s.%s",r->target->layer->name,r->target->name);
    if(r->visible_only)pos+=sprintf(pos,"!");
    if(r->x!=RULE_INTNONE)pos+=sprintf(pos,"\tx=%d%%%d",r->x,(int)(100*r->xr));
    if(r->y!=RULE_INTNONE)pos+=sprintf(pos,"\ty=%d%%%d",r->y,(int)(100*r->yr));
    if(r->w!=RULE_INTNONE)pos+=sprintf(pos,"\tw=%d%%%d",r->w,(int)(100*r->wr));
    if(r->h!=RULE_INTNONE)pos+=sprintf(pos,"\th=%d%%%d",r->h,(int)(100*r->hr));
    if(r->float_state!=RULE_INTNONE)pos+=sprintf(pos,"\tfloat-type=%s",floattype_strings[r->float_state]);
    if(r->stack_level!=RULE_INTNONE)pos+=sprintf(pos,"\tstack-level=%s",stacklvl_strings[r->stack_level]);
    if(r->no_steal)pos+=sprintf(pos,"\tno-steal");
    sockreport(buffer);
    /*
    sockreport("[%d] name=%s\trole=%s\tclassname=%s\tclass=%s -- tile=%s.%s",i,r->name,r->role,r->classN,r->classC,r->target->layer->name,r->target->name);
    */
    ++i;
  }
  return;
}

void cmd_display_get(int argc, char** argz){
  if(argc<1)return;
  if(argc==1 || !strcmp(argz[1],"geometry"))
    sockreport("geometry\t%d %d %d %d",0,0,MONITOR->mw,MONITOR->mh);
  return;
}

void cmd_hook_set(int argc, char** argz){
  if(argc<1)return;
  if(!strcmp(argz[0],"create")){
    if(hook_manage)free(hook_manage);
    if(argc==1){
      hook_manage=0;
    }else{
      hook_manage=strdup(argz[1]);
    }
    return;
  }
  if(!strcmp(argz[0],"dispose")){
    if(hook_unmanage)free(hook_unmanage);
    if(argc==1){
      hook_unmanage=0;
    }else{
      hook_unmanage=strdup(argz[1]);
    }
    return;
  }
}

void cmd_hook_list(int argc, char** argz){
  sockreport("hook create\t%s",hook_manage);
  sockreport("hook dispose\t%s",hook_unmanage);
  return;
}

void cmd_cycle(int argc, char** argz){
  if(argc<1)return;
  Client* original=0;
  if(argc==1){
    original=MONITOR->focus;
  }else{
    original=decode_client(argz[0]);
  }
  if(!original)return;
  Client* c=0;
  if(!strcmp(argz[0],"tiles")){
    c=client_cycle(original,CycleTiles);
  }else if(!strcmp(argz[0],"all")){
    c=client_cycle(original,CycleAll);
  }else if(!strcmp(argz[0],"tab-fwd")){
    c=client_cycle(original,CycleTileFwd);
  }else if(!strcmp(argz[0],"tab-back")){
    c=client_cycle(original,CycleTileBwd);
  }else{
    sockreport("unknown cycling type '%s'",argz[0]);
    return;
  }
  if(!c)return;
  xclient_activate(c);
  return;
}

void cmd_quit(int argc, char** argz){
  user_quit(0);
  return;
}

void cmd_exec(int argc, char** argz){
  if(argc<1){
    sockreport("nothing to execute!");
    return;
  }
  xorg_exec(argz[0],argz);
  return;
}

void cmd_include(int argc, char** argz){
  if(argc<1){
    sockreport("no include file given!");
    return;
  }
  FILE* incl=fopen(argz[0],"r");
  if(!incl){
    sockreport("file '%s' not found",argz[0]);
    return;
  }
  char commandbuffer[BUFSIZE];
  while(fgets(commandbuffer,BUFSIZE,incl)){
    cmd(commandbuffer);
  }
  fclose(incl);
  return;
}

void cmd_help(int argc, char** argz){
  for(int i=0;i<LENGTH(translation);i++){
    sockreport(translation[i].string);
  }
}

void cmd(const char* command){
  char* mycommand=strdup(command);

  char sep=' ';
  if(mycommand[0]==31 || mycommand[0]=='\t'){//input field separator
    sep=mycommand[0];
  }

  //handling comments
  char* comment=strchr(mycommand,'#');
  if(comment)comment[0]=0;
  //handling newlines
  char* nl=strchr(mycommand,'\n');
  if(nl)nl[0]=0;


  int count=0;
  int previouswhitespace=1;
  for(char* c=mycommand;*c;c++){
    if(*c==sep){
      previouswhitespace=1;
    }else{
      if(previouswhitespace){
	count++;
      }
      previouswhitespace=0;
    }
  }

  if(count){
    char** arguments=malloc((count+1)*sizeof(char*));
    int newcount=0;
    previouswhitespace=1;
    for(char* c=mycommand;*c;c++){
      if(*c==sep){
	previouswhitespace=1;
	*c=0;
      }else{
	if(previouswhitespace){
	  arguments[newcount]=c;
	  newcount++;
	}
	previouswhitespace=0;
      }
    }
    arguments[count]=0;

    say("> %s",arguments[0]);
    for(int i=1;i<newcount;i++){
      say("\t%s",arguments[i]);
    }
    say("\n");
    int found=0;
    for(int i=0;i<LENGTH(translation);i++){
      if(!strcmp(translation[i].string,arguments[0])){
	translation[i].fun(newcount-1,arguments+1);
	found=1;
	break;
      }
    }
    if(!found){
      say("unknown mycommand '%s'\n",arguments[0]);
    }
    free(arguments);
  }
  free(mycommand);
}
